(* Instead of using the TS-M2-internal integer representation, we switch
 * to our own type, doing the calculation in Module Ints.
 * Thus we allow for arbitrary high precisions.
 *)
MODULE Expr;

IMPORT Lib, Str, IO, Scnr, Ints;

CONST
  verStr = 'expr v4 (2024-11-03) solyga@gmx.de';

TYPE
  TValue = Ints.TValue;

VAR
  sym: Scnr.TSymbol;


PROCEDURE Version();
BEGIN
  IO.WrStr( verStr ); IO.WrLn();
END Version;

PROCEDURE Help();
BEGIN
  Version();
  IO.WrStr( 'Integer expression evaluator.' ); IO.WrLn();
  IO.WrStr( 'Usage  : "expr" [option] expression' ); IO.WrLn();
  IO.WrStr( 'Options: -v   Show version and exit.' ); IO.WrLn();
  IO.WrStr( '         -h   Show this help and exit.' ); IO.WrLn();
  IO.WrStr( '         -B   Print binary value.' ); IO.WrLn();
  IO.WrStr( '         -O   Print octal value.' ); IO.WrLn();
  IO.WrStr( '         -H   Print hexadecimal value.' );
END Help;

PROCEDURE Option( VAR base, idx: CARDINAL );
VAR
  s: ARRAY [ 0 .. 7 ] OF CHAR;
BEGIN
  base := 10;
  idx := 1;
  IF Lib.ParamCount() < 1 THEN Help(); HALT END;
  Lib.ParamStr( s, 1 );
  IF    Str.Compare( s, '-v' ) = 0 THEN
    Version(); HALT;
  ELSIF Str.Compare( s, '-h' ) = 0 THEN
    Help(); HALT;
  ELSIF Str.Compare( s, '-B' ) = 0 THEN
    base := 2; INC( idx );
  ELSIF Str.Compare( s, '-O' ) = 0 THEN
    base := 8; INC( idx );
  ELSIF Str.Compare( s, '-H' ) = 0 THEN
    base := 16; INC( idx );
  END;
END Option;

PROCEDURE Get();
BEGIN
  sym := Scnr.Read();
END Get;

(* cmd line string is stored in Scnr/Rdr to be accessable from all modules *)
PROCEDURE InitScanner( idx: CARDINAL );
TYPE
  TCmdLine = ARRAY [ 0 .. 255 ] OF CHAR;
VAR
  i, c: CARDINAL;
  s, cmdLine: TCmdLine;
BEGIN
  cmdLine := '';
  c := Lib.ParamCount();
  FOR i := idx TO c DO
    Lib.ParamStr( s, i );
    Str.Append( cmdLine, s );
    IF i < c THEN Str.Append( cmdLine, ' ' ) END;
  END;
  Scnr.Init( cmdLine );
  Get();
END InitScanner;


PROCEDURE IsAdd(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symAdd );
END IsAdd;

PROCEDURE IsSub(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symSub );
END IsSub;

PROCEDURE IsDashOp(): BOOLEAN;
BEGIN
  RETURN IsAdd() OR IsSub();
END IsDashOp;

PROCEDURE IsMul(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symMul );
END IsMul;

PROCEDURE IsDiv(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symDiv );
END IsDiv;

PROCEDURE IsDotOp(): BOOLEAN;
BEGIN
  RETURN IsMul() OR IsDiv();
END IsDotOp;

PROCEDURE IsNum(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symNum );
END IsNum;

PROCEDURE IsLBrk(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symLBrk );
END IsLBrk;

PROCEDURE IsRBrk(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symRBrk );
END IsRBrk;

PROCEDURE IsEOS(): BOOLEAN;
BEGIN
  RETURN ( sym = Scnr.symEOS );
END IsEOS;


PROCEDURE Error( str-: ARRAY OF CHAR; syms: Scnr.TSymbols );
VAR
  i: CARDINAL;
  s: Scnr.TSymbol;
  c: CARDINAL;
BEGIN
  IO.WrStr( '(' );
  IO.WrCard( Scnr.Pos(), 1 );
  IO.WrStr( ') Parser error: ' );
  IF str[0] # 0C THEN
    IO.WrStr( str );
    IO.WrStr( '.' );
  END;
  IF syms # Scnr.symsNone THEN
    IO.WrStr( 'Expected: ' );
    c := 0;
    FOR s := MIN(Scnr.TSymbol) TO MAX(Scnr.TSymbol) DO
      IF s IN syms THEN INC( c ) END;
    END;
    FOR s := MIN(Scnr.TSymbol) TO MAX(Scnr.TSymbol) DO
      IF s IN syms THEN
        IO.WrStr( Scnr.symNames[s] );
        DEC( c );
        IF c > 0 THEN IO.WrStr( ', ' ) END;
      END;
    END;
  END;
  IO.WrLn();
  Scnr.OutStr();
  (* show position -- function requires tab to be an invalid char ! *)
  FOR i := 1 TO Scnr.Pos() DO IO.WrChar( '_' ) END;
  IO.WrChar( '^' );
  IO.WrLn();
  HALT;
END Error;


(* synt = expr .
 * expr = [ "+" |"-" ] term { ( "+" | "-" ) term } .
 * term = fact { ( "*" | "/" ) fact } .
 * fact = number | "(" expr ")" .
 *)

PROCEDURE Expression( VAR val: TValue );

  PROCEDURE Factor( VAR val: TValue );
  BEGIN
    IF IsNum() THEN
      val := Scnr.Val();
      Get();
    ELSIF IsLBrk() THEN
      Get();
      Expression( val );
      IF IsRBrk() THEN
        Get();
      ELSE
        Error( '', Scnr.TSymbols{Scnr.symRBrk} );
      END;
    ELSE
      Error( '', Scnr.TSymbols{Scnr.symNum} );
    END;
  END Factor;

  PROCEDURE Term( VAR val: TValue );
  VAR x, y: TValue;
  BEGIN
    Factor( val );
    WHILE IsDotOp() DO
      x := val;
      IF IsMul() THEN
        Get();
        Factor( y );
        IF ~ Ints.Mul( val, x, y ) THEN Error( 'Multiplication overflow', Scnr.symsNone ) END;
      ELSE
        Get();
        Factor( y );
        IF ~ Ints.Div( val, x, y ) THEN Error( 'Division by zero', Scnr.symsNone ) END;
      END;
    END;
  END Term;

  PROCEDURE Update();
  VAR x, y: TValue;
  BEGIN
    x := val;
    IF IsAdd() THEN
      Get();
      Term( y );
      IF ~ Ints.Add( val, x, y ) THEN Error( 'Addition overflow', Scnr.symsNone ) END;
    ELSE
      Get();
      Term( y );
      IF ~ Ints.Sub( val, x, y ) THEN Error( 'Subtraction overflow', Scnr.symsNone ) END;
    END;
  END Update;

BEGIN
  IF IsDashOp() THEN
    val := Ints.valZero;
    Update();
  ELSE
    Term( val );
  END;
  WHILE IsDashOp() DO Update() END;
END Expression;


VAR
  val: TValue;
  str: ARRAY [ 0 .. 127 ] OF CHAR;
  ok: BOOLEAN;
  outBase: CARDINAL;
  idx: CARDINAL;
BEGIN
  Option( outBase, idx );
  InitScanner( idx );
  Expression( val );
  IF ~ IsEOS() THEN Error( '', Scnr.TSymbols{Scnr.symEOS} ) END;


  Ints.ValToStr( str, val, outBase, ok );
  IF ok THEN
    IO.WrStr( str );
    CASE outBase OF
    |  2: IO.WrStr( 'B' );
    |  8: IO.WrStr( 'O' );
    | 10:
    | 16: IO.WrStr( 'H' );
    ELSE
      Error( 'BUG: Invalid outBase', Scnr.symsNone );
    END;
    IO.WrLn();
  ELSE
    Error( 'BUG: Output conversion failed', Scnr.symsNone );
  END;
(*
  Scnr.Reset();
  Get();
  WHILE sym # Scnr.symEOS DO
    IO.WrStr( Scnr.symNames[ sym ] );
    IF sym = Scnr.symNum THEN
      IO.WrStr( ' ' );
      IO.WrCard( Scnr.Val(), 1 );
    END;
    IO.WrStr( ', ' );
    Get();
  END;
  IO.WrLn();
*)
END Expr.
